#ifndef DATA_SINGLEPATTERN_H_
#define DATA_SINGLEPATTERN_H_

#include <vector>
#include <string>
#include <stack>
#include <ext/hash_map>
#include <list>
#include <map>
#include <deque>
#include <iostream>
#include <utility>
#include "Util_String.h"
//#include <boost/shared_ptr.hpp>


using namespace std::rel_ops;
using namespace std;

    typedef vector<unsigned int>                    intVec;
    typedef vector<unsigned int>::const_iterator	intVecCITER;
    typedef pair<unsigned int,unsigned int>         intPair;
	typedef pair<intPair, intPair>                  intPPair;
	typedef vector<intPPair>::const_iterator        IntPPairCITER;   
 
//--------------------------------------------------------------------------
// class ID
//--------------------------------------------------------------------------
class ID
{
   public:
   
      ID();
      ID(const string& myMolRef,const string& myStrRef,const string& myId);
      ID(const ID &myid);
   
      virtual ~ID();
   
      const string& getMolRef()   const;
      const string& getStrRef()   const;
      const string& getId()       const;
   
   
   private:
   
      std::string molRef;
      std::string strRef;
      std::string id;
};

class IDHash 
{ 
   public:
      size_t operator()(const ID &myID) const 
      {
         unsigned long hash = 5381;
         std::string myStr = myID.getId() + myID.getMolRef() + myID.getStrRef();

         for (unsigned int i = 0; i < myStr.length(); i++) 
         {
            hash = ((hash << 5) + hash) + myStr[i]; // hash * 33 + cStateString[i]
         }

         return hash;
   } 
};

class IDEq {
public:
   bool operator()(const ID &a,const ID &b) const
   {
      return (a.getId() == b.getId());
   }
};

class StringHash 
{ 
   public:
      size_t operator()(const string &myStr) const 
      {
         unsigned long hash = 5381;
 
         for (unsigned int i = 0; i < myStr.length(); i++) 
         {
            hash = ((hash << 5) + hash) + myStr[i]; // hash * 33 + cStateString[i]
         }

         return hash;
   } 
};

class StringEq {
public:
   bool operator()(const string &a,const string &b) const
   {
      return (a == b);
   }
};

//--------------------------------------------------------------------------
// class SinglePattern 
//     this is part of the pattern in one structure 
//--------------------------------------------------------------------------
class SinglePattern
{
public:
      
      SinglePattern();
      SinglePattern(const SinglePattern& mySingP);
      SinglePattern(const string& myMolRef,const string& myStrRef,const string& myId,const intVec& mySinglePattern);
 
	virtual ~SinglePattern();
   
   const ID&            getID()  const;
   const intVec&        getPat() const;
         void           print();
   
private:

      ID          Id;
      intVec         pattern;   
}; 

//--------------------------------------------------------------------------
// class Pattern
//    this is the hole pattern with parts in different structure
//    therefore a Pattern consits of SinglePatterns 
//--------------------------------------------------------------------------
/*class Pattern
{
   public:
   
      typedef __gnu_cxx::hash_map<ID,SinglePattern,IDHash,IDEq> selfMapTYPE;
      
               Pattern();
               Pattern(const Pattern& myP);
               Pattern(const SinglePattern& mySinglePat);
      virtual  ~Pattern();
      
      const    string&        getid()                    const;
      const    SinglePattern& getSinglePattern(const ID& myID)  const;
      const    selfMapTYPE&   getMap()                   const;
               void           add(const SinglePattern& myPat);
      
   private:
     
     string       id;
     selfMapTYPE  pattern;
};*/

//--------------------------------------------------------------------------
// class PatternMap
//    is able to manage an arbitrary bunch of patterns 
//--------------------------------------------------------------------------
/*class PatternMap
{
   public:
      
      typedef __gnu_cxx::hash_map<string,Pattern,StringHash,StringEq> selfMapTYPE;
   
               PatternMap();
      virtual ~PatternMap();
      
      const Pattern&       getPattern(const string& id) const;
            void           add(const SinglePattern& myPat);
            void           add(const Pattern& myPat);
      const selfMapTYPE&   getMap() const;
      const int            size();
      
   private:
   
     selfMapTYPE patternMap;
};
*/
//--------------------------------------------------------------------------
// class PatternPairMap
//    is able to manage a pattern with 2 singlepatterns from pairwise_mcs 
//--------------------------------------------------------------------------
class PatternPair
   {
      public:
      PatternPair(){};
      PatternPair(const PatternPair &myPair)
                     :id(myPair.id),size(myPair.size),first(myPair.first),second(myPair.second)
      {
    	  score = 0;
      };
            
      PatternPair(const string& myId,const int& mySize,const SinglePattern& myFirstPat,const SinglePattern& mySecPat)
                  : id(myId),size(mySize),first(myFirstPat),second(mySecPat)
      { 
    	  score = 0;
      };
      
      const string& getId() const
      {
         return id;
      };
      
      const int& getSize() const
      {
         return size;
      };
      
      const SinglePattern& getFirstPat() const
      {
         return first;
      };
      
      const SinglePattern& getSecPat() const
      {
         return second;
      };

      void resetBounds()
      {
    	  insideBounds.clear();
      }
      
      void setOutsideBounds(intPPair myPPair)
      {
    	  outsideBounds = myPPair;
      };
      
      intPPair getOutsideBounds()
      {
    	  return outsideBounds;
      };
      
      void addInsideBounds(intPPair myPPair)
      {
    	  insideBounds.push_back(myPPair);
      };
      
      const vector<intPPair>& getInsideBounds()
      {
    	  return insideBounds;
      };
      
      void setEPMScore(int myScore)
      {
    	  score = myScore;
      };
      
      int getScore()
            {
          	  return score;
            };
            
        
      private:
         string         	id;
         int            	size; 
         SinglePattern  	first;
         SinglePattern  	second;
         
         int				score;
         vector<intPPair>   insideBounds;
         intPPair           outsideBounds;
   };



//--------------------------------------------------------------------------
// class PatternPairMap
//    to manage a pair of 2 single patterns like they are generated by
//    pairwise_mcs/calculatePatterns 
//--------------------------------------------------------------------------
class PatternPairMap
{     
   public:
      typedef  PatternPair                                  selfValueTYPE;
      //typedef  boost::shared_ptr<PatternPair>             SelfValuePTR;
      typedef  PatternPair*				               		SelfValuePTR;
      
      typedef  multimap<int,SelfValuePTR,greater<int> >     orderedMapTYPE;
      typedef  orderedMapTYPE::const_iterator               orderedMapCITER;
      typedef  orderedMapTYPE::iterator                     orderedMapITER;
      typedef  list<SelfValuePTR>                           patListTYPE;
      typedef  patListTYPE::iterator                        patListITER;
      typedef  patListTYPE::const_iterator                  patListCITER;
      typedef  __gnu_cxx::hash_map<string,SelfValuePTR,StringHash,StringEq> PatternIdMapTYPE;
   
         PatternPairMap();
         PatternPairMap(const PatternPairMap& myPairMap)
                           :patternList(myPairMap.patternList),
                            patternOrderedMap(myPairMap.patternOrderedMap),
                            idMap(myPairMap.idMap)  {};
      
      virtual ~PatternPairMap();
       
               void              add( const string& id,
                                      const int& mysize,
                                      const SinglePattern& first,
                                      const SinglePattern& second);
               void              add(const SelfValuePTR value);
               void              makeOrderedMap();
               void              updateFromMap();
      const    PatternPair&      getPatternPair(const string& id)const;
      const    SelfValuePTR      getPatternPairPTR(const string& id)const;
      const    patListTYPE&      getList() const;
      const    orderedMapTYPE&   getOrderedMap() const;
               orderedMapTYPE&   getOrderedMap2();         
      const    int               size()   const;
      
   private:
   
     patListTYPE        patternList;
     orderedMapTYPE     patternOrderedMap;
     PatternIdMapTYPE   idMap;
};




//--------------------------------------------------------------------------
// class SinglePatternMap
//    is able to manage an arbitrary bunch of single patterns
//    key is Id
//--------------------------------------------------------------------------
/*class SinglePatternMap
{
   public:

      typedef __gnu_cxx::hash_map<string,SinglePattern,StringHash,StringEq>   selfMapTYPE;
      typedef selfMapTYPE::const_iterator   selfMapCITER;
         
               SinglePatternMap();
      virtual  ~SinglePatternMap();
      
      const    SinglePattern  getSinglePattern(const string& id) const;
               void           add(const SinglePattern& mySingPat);
               //void           add(SinglePattern* mySingPat);
      
      const    selfMapTYPE&   getMap();
      
   private:
   
               selfMapTYPE    singlePatternMap;
};
*/
//--------------------------------------------------------------------------
// class SimplePatternMMap
//    is a multimap of SinglePatterns, key is patternsize 
//--------------------------------------------------------------------------
/*class SimplePatternMap
{
   public:
      typedef multimap<int,SinglePattern,greater<int> >  selfMapTYPE;
      typedef selfMapTYPE::const_iterator                selfMapITER;
      
               SimplePatternMap();
      virtual  ~SimplePatternMap();
      
    //  const    SinglePattern  getSinglePattern(string id) const;
               void           add(const SinglePattern& mySingPat);
      
      const    selfMapTYPE&   getMap();
      
   private:
   
               selfMapTYPE    simplePatternMap;
};
*/
//--------------------------------------------------------------------------
// class Molecule
//       stores all information for one molecule/sequence 
//--------------------------------------------------------------------------
class Molecule
{ 
   
   public:
                           Molecule();
                           Molecule(const Molecule& myMol);
                           Molecule(const string& myMolId, 
                                    const string& mySeq,
                                    const string& myStrId,
                                    const string& myStr);     
      virtual             ~Molecule();
      
      const string&        getId()                                      const;
      const string&        getSequence()                                const;
      const char           getBase(const unsigned int& position)        const;
      const string&        getStructure()                               const;
      const string&        getActualStructureId()                       const;
      const char           getStructure(const unsigned int& position)   const;
      const vector<int>&   getBasePairs()                               const; 
      const int            getBasePartner(const unsigned int& position) const;
      const int            getLength()                                  const;   
            
            void           setActualStructureId(const string& strId);
            void           printInfo(const bool& shortview);

   private:
   
      typedef __gnu_cxx::hash_map <string,string,StringHash,StringEq>      MAPTYPE_Structures;
      typedef __gnu_cxx::hash_map <string,vector<int>,StringHash,StringEq> MAPTYPE_BasePairs;
      
            string               id;
            string               sequence;
            MAPTYPE_Structures   structures;
            MAPTYPE_BasePairs    basePairs;
            string               actual_structure;   
            void                 makeBasePairs(string myStr);   
};

#endif /*DATA_SINGLEPATTERN_H_*/
